import glob
import os
import shutil
import platform
import subprocess
from pathlib import Path
import sysconfig
import sys

from hatchling.builders.hooks.plugin.interface import BuildHookInterface

def clean_folder(folder):
    files = glob.glob(folder + "/*")
    for f in files:
        if os.path.isfile(f):
            os.remove(f)
        else:
            if not (os.path.basename(f) == "Mac" or os.path.basename(f) == "Win"):
                shutil.rmtree(f)

def copy_files(fromPath, toPath, files, bads = [], bad_word = ""):
    for f in files:
        if f not in bads and (bad_word == "" or bad_word not in f):
            shutil.copy(fromPath + "/" + f, toPath)

def glob_and_copy(from_path, to_path, pattern):
    files_globbed = glob.glob(pattern, root_dir=from_path)
    copy_files(from_path, to_path, files_globbed)

class CustomBuildHook(BuildHookInterface):
    def initialize(self, version, build_data):
        platform_config = sysconfig.get_platform()
        dlls_path = "dymo_sdk/dlls"
        curr_file = Path(self.directory).parent.resolve()
        if "platform" in os.environ and os.environ["platform"] != "":
            chosen_plat = os.environ["platform"]
        else:
            chosen_plat = platform.system()
        if "arch" in os.environ and os.environ["arch"] != "":
            arch = os.environ["arch"]
        else:
            arch = ""
        
        #If we're calling from the build agent, the build agent will automatically build on it's own
        skip_dot_net_build = ("run_from_agent" in os.environ and os.environ["run_from_agent"] == "true")
        
        if chosen_plat == "Windows":
            #Windows
            #path to projects for building
            successful = True
            sdk_dlls = curr_file.parent / "DymoSDK" / "bin" / "Release" / "net8.0-windows10.0.19041.0"
            sdk_dlls = str(sdk_dlls.resolve())

            label_api_dlls = curr_file.parent.parent / "Components" / "DYMO.LabelAPI.Windows" / "bin" / "x64" / "Release" / "net8.0-windows10.0.19041.0" / "win-x64"
            label_api_dlls = str(label_api_dlls.resolve())


            if not skip_dot_net_build:
                sdk_path = curr_file.parent / "DymoSDK" / "DymoSDK.csproj"
                sdk_path = str(sdk_path.resolve())
                label_api_win_path = curr_file.parent.parent / "Components" / "DYMO.LabelAPI.Windows" / "DYMO.LabelAPI.Windows.csproj"
                label_api_win_path = str(label_api_win_path.resolve())

                inter_folder1 = ""
                inter_folder2 = ""
                if arch == "":
                    if "arm64" in platform_config:
                        inter_folder1 = "arm64"
                        inter_folder2 = "win-arm64"
                    elif "amd64" in platform_config:
                        inter_folder1 = "x64"
                        inter_folder2 = "win-x64"
                    elif "win32" in platform_config:
                        inter_folder1 = "x86"
                        inter_folder2 = "win-x86"
                else:
                    if arch == "arm64":
                        inter_folder1 = "arm64"
                        inter_folder2 = "win-arm64"
                    elif arch == "x64":
                        inter_folder1 = "x64"
                        inter_folder2 = "win-x64"
                    elif arch == "x86":
                        inter_folder1 = "x86"
                        inter_folder2 = "win-x86"

                #path to dlls
                bin_path_sdk = curr_file.parent / "DymoSDK" / "bin" / "Release" / "net8.0-windows10.0.19041.0"
                bin_path_sdk = str(bin_path_sdk.resolve())
                bin_path_label_api = curr_file.parent.parent / "Components" / "DYMO.LabelAPI.Windows" /"bin" / inter_folder1 / "Release" / "net8.0-windows10.0.19041.0" / inter_folder2
                bin_path_label_api = str(bin_path_label_api.resolve())

                clean_command = ["dotnet", "clean", sdk_path]
                build_command = ["dotnet", "build", sdk_path, "-c", "Release"]
                clean_command_win = ["dotnet", "clean", label_api_win_path]
                build_command_win = ["dotnet", "build", label_api_win_path, "-c",  "Release", "--arch", inter_folder1]

                try:
                    dotnet_clean_result = subprocess.run(clean_command, check=True, capture_output=True, text=True, shell=True)
                    dotnet_build_result = subprocess.run(build_command, check=True, capture_output=True, text=True, shell=True)
                    dotnet_clean_result_win = subprocess.run(clean_command_win, check=True, capture_output=True, text=True, shell=True)
                    dotnet_build_result_win = subprocess.run(build_command_win, check=True, capture_output=True, text=True, shell=True)
                except subprocess.CalledProcessError as e:
                    print(f"Ran into error: {e}")
                    print(e.stdout)
                    print("now printing errors")
                    print(e.stderr)
                    successful = False
                    #raise e
            
            #copy the required files
            if successful:
                clean_folder(dlls_path)
                glob_and_copy(sdk_dlls, dlls_path, "*.dll")
                glob_and_copy(label_api_dlls, dlls_path, "*.dll")
                print("successfully copied dlls")
            else:
                print("did not copy new dlls")
        else:
            #Mac
            successful = True
            sdk_path = curr_file.parent / "DymoSDK" / "DymoSDK.csproj"
            sdk_path = str(sdk_path.resolve())
            
            label_api_mac_path = curr_file.parent.parent / "Components" / "DYMO.LabelAPI.Mac" / "DYMO.LabelAPI.Mac.csproj"
            label_api_mac_path = str(label_api_mac_path.resolve())

            bin_path_sdk = curr_file.parent / "DymoSDK" / "bin" / "Release" / "netstandard2.0"
            bin_path_sdk = str(bin_path_sdk.resolve())
            bin_path_label_api = curr_file.parent.parent / "Components" / "DYMO.LabelAPI.Mac" / "bin" / "Release" / "net8.0-macos"
            bin_path_label_api = str(bin_path_label_api.resolve())

            if not skip_dot_net_build:
                clean_command = ["dotnet clean " + sdk_path]
                build_command = ["dotnet build " + sdk_path + " --configuration Release"]
                clean_command_2 = ["dotnet clean " + label_api_mac_path]
                build_command_2 = ["dotnet build " + label_api_mac_path + " --configuration Release"]
                try:
                    dotnet_clean_result = subprocess.run(clean_command, check=True, capture_output=True, text=True, shell=True)
                    dotnet_build_result = subprocess.run(build_command, check=True, capture_output=True, text=True, shell=True)
                    dotnet_clean_result_2 = subprocess.run(clean_command_2, check=True, capture_output=True, text=True, shell=True)
                    dotnet_build_result_2 = subprocess.run(build_command_2, check=True, capture_output=True, text=True, shell=True)
                except subprocess.CalledProcessError as e:
                    print(f"Ran into error: {e}")
                    print(e.stdout)
                    print("now printing errors")
                    print(e.stderr)
                    successful = False
                    #raise e
            
            #copy the required files
            if platform.system() == "Windows":
                native_build_command = ["./build.native.osx.arm64.sh"]
            else:
                native_build_command = ["source build.native.osx.arm64.sh"]
            try:
                build_native_result = subprocess.run(native_build_command, check=True, capture_output=True, text=True, shell=True, cwd="../../Components.Native")
            except subprocess.CalledProcessError as e:
                    print(f"Ran into error: {e}")
                    print(e.stdout)
                    print("now printing errors")
                    print(e.stderr)
            if successful:
                clean_folder(dlls_path)
                glob_and_copy(bin_path_sdk, dlls_path, "*.dll")
                glob_and_copy(bin_path_sdk, dlls_path, "*.a")
                glob_and_copy(bin_path_sdk, dlls_path, "*.dylib")

                glob_and_copy(label_api_mac_path, dlls_path, "*.dll")
                glob_and_copy(label_api_mac_path, dlls_path, "*.a")
                glob_and_copy(label_api_mac_path, dlls_path, "*.dylib")

                native_path = "../../native/osx/dylib/Release"
                glob_and_copy(native_path, dlls_path, "*.dylib")
                mac_os_native_skiasharp = str(Path.home().resolve()) + "/.nuget/packages/skiasharp.nativeassets.macos/3.116.1/runtimes/osx/native/libSkiaSharp.dylib"
                os.makedirs(dlls_path + "/@rpath/libSkiaSharp.framework", exist_ok=True)
                #copy to both original folder, and rpath folder since sometimes it looks at one or the other
                shutil.copy(mac_os_native_skiasharp, dlls_path)
                shutil.copy(mac_os_native_skiasharp, dlls_path + "/@rpath/libSkiaSharp.framework")

                #grab framework file
                if platform.system() == "Windows":
                    #hope 9.0 works, else we will need to install 8.0 onto our build agent
                    mac_os_dll_path - "Program Files/dotnet/packs/Micrososft.macOS.Runtime.osx-arm64.net9.0_15.0/15.5.9219/runtimes/osx-arm64/lib/net9.0/Microsoft.macOS.dll"
                else:
                    mac_os_dll_path = "/usr/local/share/dotnet/packs/Microsoft.macOS.Runtime.osx-arm64.net8.0_15.0/15.0.8319/runtimes/osx-arm64/lib/net8.0/Microsoft.macOS.dll"
                
                shutil.copy(mac_os_dll_path, dlls_path)

                print("successfully copied dlls")


        #set build tags
        build_data["infer_tag"] = False

        build_data["pure_python"] = False

        if platform.system() == "Windows":
            platform_tag = platform_config.replace("-", "_").replace(".","_")
            if arch == "arm64":
                platform_tag = platform_tag.replace("amd64", "arm64")
            elif arch == "x86":
                platform_tag = platform_tag.replace("win_amd64", "win32")
            print(f"platform_tag:{platform_tag}")
        else:
            platform_tag = "macosx_13_0_arm64"
        
        abi_tag = "none"
        python_tag = "py2.py3"

        build_data["tag"] = python_tag + "-" + abi_tag + "-" + platform_tag


from .core import *

__all__ = [
    "LabelObject",
    "DymoLabel",
    "DymoPrinter",
    "RollStatus",
    "get_printers",
    "print_label",
    "is_roll_status_supported",
    "refresh_printer",
    "get_roll_status"
]



"""Provides functionality for interacting with Dymo Labels

This package provides a python wrapper around the .NET SDK for DYMO Printers and Labels.
"""
import pythonnet
import os
from pythonnet import load
import clr_loader
import asyncio
from packaging.version import Version
from enum import Enum
from enum import auto
import sys
import shutil
import glob
import platform

__all__ = [
    "LabelObject",
    "DymoLabel",
    "DymoPrinter",
    "RollStatus",
    "get_printers",
    "print_label",
    "is_roll_status_supported",
    "refresh_printer",
    "get_roll_status"
]

#load the Desktop .Net core to help ensure functionality
__specToLoad = ""
__spec_ver = "0"
for r in clr_loader.find_runtimes():
    if platform.system() == "Windows":
        if r.name == "Microsoft.WindowsDesktop.App" and Version(r.version) >= Version(__spec_ver):
            __specToLoad = r
            __spec_ver = r.version
    else:
        if r.name == "Microsoft.NETCore.App" and r.version[0] == "9":
        if r.name == "Microsoft.NETCore.App" and r.version[0] == "8":
            __specToLoad = r
            __spec_ver = r.version
if __specToLoad != "":
    clr_instance = clr_loader.get_coreclr(runtime_spec=__specToLoad)
    load(clr_instance)
else:
    load("coreclr")

import clr
import System

__curr_dir = os.path.dirname(os.path.abspath(__file__))
__dlls_dir = os.path.join((__curr_dir), "dlls")

#This section is to copy the appropriate files
#Note: You can remove this code section in the actual package
#if platform.system() == "Windows":
#    __p_dir = root_dir=os.path.join((__curr_dir), "dlls/Win")
#else: #if not windows, add mac and hope it's compatible with windows
#    __p_dir = root_dir=os.path.join((__curr_dir), "dlls/Mac")
#__dlls = glob.glob("*", root_dir=__p_dir)
#for __dll in __dlls:
#    if os.path.isdir(__p_dir + "/" + __dll):
#        if os.path.exists(__dlls_dir + "/" + __dll):
#            shutil.rmtree(__dlls_dir + "/" + __dll)
#        shutil.copytree(__p_dir + "/" + __dll, __dlls_dir + "/" + __dll)
#    else:    
#        shutil.copy(__p_dir + "/" + __dll, __dlls_dir)

sys.path.append(__dlls_dir)

clr.AddReference('DymoSDK')
clr.AddReference('Dymo.LabelAPI')

#to allow debugger to attach if needed
#__test = input("press enter to continue")

import DymoSDK

from DymoSDK.Implementations import DymoPrinter as NativeDymoPrinter
from DymoSDK.Implementations import DymoLabel as NativeDymoLabel
from DymoSDK import App
from System import Convert
import DYMO.LabelAPI
from DymoSDK.Interfaces import IPrinter
App.Init() #where should this be run?
__printer_manager = NativeDymoPrinter.Instance

class LabelObjectType(Enum):
    """An ENUM representing the type of object."""
    TEXT = auto()
    ADDRESS = auto()
    COUNTER = auto()
    BARCODE = auto()
    QRCODE = auto()
    IMAGE = auto()
    #Types that cannot be updated
    SHAPE = auto()
    NONE = auto()

class LabelObject:
    """An object on a label, such as Text, Barcode, etc. Should not be explicitly constructed.
    
    Attributes:
        native: The native C# label object, should not be needed for the average user.
        owner: The DymoLabel instance that this object is attached to.
        object_type: The type of object this label is.
    """
    _typeDict = {
        "TextObject" : LabelObjectType.TEXT,
        "AddressObject" : LabelObjectType.ADDRESS,
        "CounterObject" : LabelObjectType.COUNTER,
        "BarcodeObject" : LabelObjectType.BARCODE,
        "QRCodeObject" : LabelObjectType.QRCODE,
        "ImageObject" : LabelObjectType.IMAGE,
        "ShapeObject" : LabelObjectType.SHAPE,
        "None" : LabelObjectType.NONE
    } 

    def __init__(self, native_label_object, label):
        """Creates a Python LabelObject from a C# LabelObject."""
        self.native = native_label_object
        self.object_type = LabelObject._typeDict[native_label_object.Type.ToString()]
        self.owner = label
    
    @property
    def name(self):
        """The name of the label object."""
        return self.native.Name
    @name.setter
    def name(self, new_name):
        self.owner._nameToPythonMap[new_name] = self
        self.owner._nameToPythonMap.pop(self.native.Name, None)
        self.native.Name = new_name
    
    
    def update_data(self, new_value: str):
        """Updates the current data using the new_value as a String."""
        return self.owner.native.UpdateLabelObject(self.native, new_value)
    
    def update_image_data_from_file(self, filepath: str):
        """If this object is an image object, updates the image data using the file provided."""
        if self.object_type != LabelObjectType.IMAGE:
            raise TypeError("You cannot call update_image_data_from_file on a non image object.")
        
        return self.owner.native.SetImageFromFilePath(self.name, filepath)


    
    

class DymoLabel:
    """A Dymo Label object.
    
    Attributes:
        :native: The underlying DymoLabel object from C#.
    """

    def __init__(self, *, filepath: str ="", xml: str =""):
        """Initializes the Label Object, optionally from a file or XML string.

        Args:
            :filepath (str): The path to a .dymo or .label file to load the label from.
            :xml (str): An XML string to load the label from.
        """

        #WARNING: This is deprecated and will need to be changed in the future
        self.native = NativeDymoLabel.Instance
        self._nameToPythonMap = {}
    
        if filepath != "":
            self.native.LoadLabelFromFilePath(filepath)
        elif xml != "":
            utf8Bom = '\ufeff'
            if xml.startswith(utf8Bom):
                xml = xml[len(utf8Bom):]
            self.native.LoadLabelFromXML(xml)
    
    def load_label_from_xml(self, xml: str):
        """Loads a label from an XML String.

        Args:
            :xml (str): An XML String describing a Dymo label.
        """
        utf8Bom = '\ufeff'
        if xml.startswith(utf8Bom):
            xml = xml[len(utf8Bom):]
        self.native.LoadLabelFromXML(xml)
    
    def load_label_from_file(self, filepath: str):
        """Loads a label from a .dymo, .label, or .xml file.

        Args:
            :filepath (str): the path to the file
        """
        self.native.LoadLabelFromFilePath(filepath)

    def get_label_objects(self):
        """Returns a list of LabelObjects on the Label."""
        result_list = []
        nativeLabelObjects = self.native.GetLabelObjects()
        for lo in nativeLabelObjects:
            if lo.Name in self._nameToPythonMap:
                result_list.append(self._nameToPythonMap[lo.Name])
            else:
                new_label_object = LabelObject(lo, self)
                self._nameToPythonMap[new_label_object.name] = new_label_object
                result_list.append(new_label_object)
        
        return result_list
    
    def get_label_object(self, obj_name: str):
        """Returns the Label Object from the given obj_name, or none if it does not exist."""
        label_obj = self.native.GetLabelObject(obj_name)
        if label_obj == None:
            return None
        if label_obj.Name in self._nameToPythonMap:
            return self._nameToPythonMap[label_obj.Name]
        new_label_obj = LabelObject(label_obj, self)
        return new_label_obj

    def update_label_object(self, label_obj: LabelObject, new_value: str):
        """Updates the given label object on this label with the new value.

        Updates the label object of this label if found. If label_obj's owner is the same as this current label, then provides duplicate functionality of label_obj.update_data().
        """
        return self.native.UpdateLabelObject(label_obj.native, new_value)
    
    def get_preview_label(self) -> str:
        """Returns a preview of the current label.
        
        Returns the current label as an image by a base64 string. 
        """
        return Convert.ToBase64String(self.native.GetPreviewLabel())
    
    def update_image_object_from_file(self, label_obj: LabelObject | str, filepath: str):
        """Updates an image object using the given filepath.

        Args:
            :label_obj: Either LabelObject Instance or the name of such an instance.
            :filepath (str): the path to the file.
        """
        if type(label_obj) == LabelObject:
            return self.native.SetImageFromFilePath(label_obj.name, filepath)
        elif type(label_obj) == str:
            return self.native.SetImageFromFilePath(label_obj, filepath)
        else:
            raise TypeError("label_obj must be a string or LabelObject instance!")
    
    def save(self, filepath: str):
        """Saves this label to the specified file path."""
        self.native.Save(filepath)


class RollStatus:
    """A class that represents the roll status.

    Attributes:
        :sku: (str) the SKU of the roll.
        :name: (str) the name of the roll.
        :labels_remaining: (int) the number of labels remaining in the detected roll.
        :RollStatus: (str) The status of the roll.
    """
    def __init__(self, native):
        self.native = native
        self.sku = native.SKU
        self.name = native.Name
        self.labels_remaining = native.LabelsRemaining
        self.roll_status = native.RollStatus

class DymoPrinter:
    """A class representing Dymo Printers. Should not be constructed directly but instead found through get_printers()."""
    _name_to_python_dict = {}
    __printer_manager = NativeDymoPrinter.Instance
    def __init__(self, native):
        """Initializes a dymo printer."""
        self.native = native
    @property
    def name(self):
        return self.native.Name
    @property
    def driver_name(self):
        return self.native.DriverName
    @property
    def is_twin_turbo(self):
        return self.native.IsTwinTurbo
    @property
    def is_local(self):
        return self.native.IsLocal
    @property
    def is_connected(self):
        return self.native.IsConnected
    @property
    def is_auto_cut_supported(self):
        return self.native.IsAutoCutSupported
    @property
    def printer_type(self):
        return self.native.PrinterType
    
    def is_roll_status_supported(self) -> bool:
        """Returns if this printer belongs to the 550 series."""
        return DymoPrinter.__printer_manager.IsRollStatusSupported(self.name)
    
    def get_roll_status(self) -> RollStatus:
        """Returns roll status."""
        return RollStatus(DymoPrinter.__printer_manager.GetRollStatusInPrinter(self.name).Result)

    def print_label(self, label: DymoLabel, copies: int = 1, collate: bool = False, mirror: bool = False, roll_selected: int = 0, chain_marks: bool = False, barcode_graphics_quality: bool = False):
        """Prints the given label on this printer

        Args:
            :label: A DymoLabel instance to print.
            :copies: The number of copies to print.
            :collate: Whether or not the print job should be collated.
            :mirror: Whether or not the print job should be mirrored.
            :roll_selected: On Twin Turbo 450 Printers you can specify which port to print out of. 0 is auto, and 1 is left, and 2 is right.
            :chain_marks: On D1 Printers places markings between labels.
            :barcode_graphics_quality: Enables higher quality printing to help ensure barcode quality. Can lead to slower print jobs.
        """
    
        return DymoPrinter.__printer_manager.PrintLabel(label.native, self.name, copies, collate, mirror, roll_selected, chain_marks, barcode_graphics_quality)

    def refresh_connection(self):
        """Refreshes the printer connection. 
        
        Note that this is done asynchronously and we recommend sleeping for 0.1 seconds if the results are required immediately.
        """
        DymoPrinter.__printer_manager.GetPrinterStatus(self.name)

    
def get_printers() -> list[DymoPrinter]:
    """Returns a list of DymoPrinter objects."""
    printers = __printer_manager.GetPrinters()
    printers = System.Collections.Generic.List[IPrinter](printers)
    res = []
    for p in printers:
        #if(p.Name in DymoPrinter._name_to_python_dict):
        #    new_p = DymoPrinter(p.Name, p.DriverName, p.IsTwinTurbo, p.IsLocal, p.IsConnected, p.IsAutoCutSupported, p.PrinterType)
        #    DymoPrinter._name_to_python_dict[p.Name] = new_p
        #    res.append(DymoPrinter._name_to_python_dict[p.Name])
        #else:
        new_p = DymoPrinter(p)
        DymoPrinter._name_to_python_dict[p.Name] = new_p
        res.append(new_p)
    return res

def print_label(label: DymoLabel, printer: DymoPrinter | str, copies: int = 1, collate: bool = False, mirror: bool = False, roll_selected: int = 0, chain_marks: bool = False, barcode_graphics_quality: bool = False):
    """Prints the given label on the specified printer asynchronously.
    
    Args:
        :label: A DymoLabel instance to print.
        :printer: The DymoPrinter to print it from. Can be a DymoPrinter instance or just the name.
        :copies: The number of copies to print.
        :collate: Whether or not the print job should be collated.
        :mirror: Whether or not the print job should be mirrored.
        :roll_selected: On Twin Turbo 450 Printers you can specify which port to print out of. 0 is auto, and 1 is left, and 2 is right.
        :chain_marks: On D1 Printers places markings between labels.
        :barcode_graphics_quality: Enables higher quality printing to help ensure barcode quality. Can lead to slower print jobs.
    """
    if type(printer) is DymoPrinter:
        printer = printer.name
    
    return __printer_manager.PrintLabel(label.native, printer, copies, collate, mirror, roll_selected, chain_marks, barcode_graphics_quality)

def is_roll_status_supported(printer_name: str) -> bool:
    """Returns if the provided printer belongs to the 550 series.
    
    Args:
        :printer_name (str): The name of the printer.
    """
    return __printer_manager.IsRollStatusSupported(printer_name)

def get_roll_status(printer_name : str) -> RollStatus:
    """Returns the number of labels remaining.

    Returns the number of labels remaining on the roll if rollstatus is supported. To update this number you must run refresh_connection() or refresh_printer().

    Args:
        :printer_name (str): The name of the printer.
    """
    return RollStatus(__printer_manager.GetRollStatusInPrinter(printer_name).Result)

def refresh_printer(printer_name: str):
    """Refreshes the printer connection
    
    Note that this is done asynchronously and we recommend sleeping for a small amount of time if the results are required immediately."""
    __printer_manager.GetPrinterStatus(printer_name)
